/*
 *	JSCWindow
 *	(SwingOSC classes for SuperCollider)
 *
 *	A replacement for (Cocoa) SCWindow.
 *
 *	Not yet working:
 *	- bounds are not tracked
 *
 *	Different behaviour
 *	- bounds are automatically restored when quitting
 *	  minimization or full screen (cocoa windows don't do this)
 *
 *	Added features
 *	- method id returns the node ID
 *
 *	@author		SuperCollider Developers
 *	@author		Hanns Holger Rutz
 *	@version		0.43, 28-Oct-06
 */
JSCWindow : Object
{
	classvar <>verbose = false;

	classvar <>allWindows;
	
	var dataptr, <name, <>onClose, <view, <userCanClose=true;
	var <alwaysOnTop=false;
	var <drawHook;
	var <acceptsMouseOver=false;
	
	// JJJ begin
	var <server, id;
	var bounds;
	var acResp;	// OSCresponderNode for window listening
	var penID		= nil;
	var <visible	= false;
	// JJJ end
	
	*initClass {
		UI.registerForShutdown({ this.closeAll });
	}

	*new { arg name = "panel", bounds, resizable = true, border = true, server;
		^super.new.initSCWindow( name, bounds, resizable, border, server );
	}
	
	// JJJ begin
	drawHook_ { arg func;
		if( drawHook.isNil, {
			if( func.notNil, {
				penID	= server.nextNodeID;
				server.sendBundle( nil,
					[ '/local', penID, '[', '/new', 'de.sciss.swingosc.Pen', '[', '/ref', this.id, ']', ']' ],
					[ '/method', this.id, \setIcon, '[', '/ref', penID, ']' ]
				);
			});
		}, {
			if( func.isNil, {
				server.sendBundle( nil,
					[ '/method', this.id, \setIcon, '[', '/ref', \null, ']' ],
					[ '/method', penID, \dispose ],
					[ '/free', penID ]
				);
				penID = nil;
			});
		});
		drawHook = func;
	}
	// JJJ end

	/*
	 *	@param	argName		is mapped to property 'title'
	 *	@param	resizable		is mapped to property 'resizable'
	 *	@param	border		is mapped to property 'undecorated'
	 *	@todo	argBounds 	are translated to java's coordinate system
	 */
	initSCWindow { arg argName, argBounds, resizable, border, argServer;
		name			= argName.asString;
		argBounds		= argBounds ?? { Rect.new( 128, 64, 400, 400 )};
		server		= argServer ?? { SwingOSC.default; };
		allWindows	= allWindows.add( this );
		id			= server.nextNodeID;
		dataptr		= this.id;
								// parent, bounds
//		view			= JSCTopView( nil, argBounds.moveTo( 0, 0 ), server );
//		id			= view.id;
		this.prInit( name, argBounds, resizable, border ); // , view );
	}

	asView { ^view }
	add { arg aView; view.add( aView )}
	
	*closeAll {
		var list;
		list = allWindows.copy;
		allWindows = Array.new(8);
		list.do({ arg window; window.close; });
	}
		
	close {
		this.prClose;
	}
	
	closed {
		dataptr = nil;
		view.prClose;
		onClose.value; // call user function
		allWindows.remove( this );
	}
	
	isClosed { ^dataptr.isNil; }
	
	// JJJ begin
	visible_ { arg boo;
		visible = boo;
		server.sendBundle( nil,
			[ '/set', this.id, \visible, boo ]);
	}	
	// JJJ end

	fullScreen {
		server.sendMsg( '/set', this.id, 'graphicsConfiguration.device.fullScreenWindow', '[', '/ref', this.id, ']' );
//		 _SCWindow_BeginFullScreen
// 		^this.primitiveFailed
	}
	
	endFullScreen {
		server.sendMsg( '/set', this.id, 'graphicsConfiguration.device.fullScreenWindow', '[', '/ref', \null, ']' );
//		"JSCWindow.endFullScreen : not yet implemented".error;
//		_SCWindow_EndFullScreen
//		^this.primitiveFailed
	}
	
	userCanClose_ { arg boo;
		userCanClose			= boo;
												// HIDE_ON_CLOSE, DO_NOTHING_ON_CLOSE
		server.sendMsg( '/set', this.id, \defaultCloseOperation, if( boo, 1, 0 ) );
//		_SCWindow_SetShouldClose
//		^this.primitiveFailed
	}
	
	acceptsMouseOver_{arg bool;
		acceptsMouseOver = bool;	
		this.prSetAcceptMouseOver(bool);
	}	

	front {
		if( drawHook.notNil, { this.refresh; });
		server.sendBundle( nil,
			[ '/set', this.id, \visible, true ],
			[ '/method', this.id, \toFront ]);
		visible = true;
//		_SCWindow_ToFront
//		^this.primitiveFailed
	}
	 
	alwaysOnTop_{ arg boolean = true;
		alwaysOnTop = boolean;
		this.prSetAlwaysOnTop( boolean );
	}
	
	prSetAlwaysOnTop{ arg boolean = true;
		server.sendMsg( '/set', this.id, \alwaysOnTop, boolean );
//		if( boolean && verbose, { "JSCWindow.alwaysOnTop_( true ) : has no effect".error; });
//		_SCWindow_AlwaysOnTop	
	}
	
	refresh {
		if( drawHook.isNil, {
			server.sendMsg( '/method', this.id, \repaint );
		}, {
			JPen.protRefresh(drawHook, view, server, penID, this.id );
		});
//		^view.refresh;
//		server.sendBundle( nil, [ '/method', this.id, \revalidate ], [ '/method', this.id, \repaint ]);
//		_SCWindow_Refresh
//		^this.primitiveFailed
	}
	
	minimize {
		// java.awt.Frame.ICONIFIED
		server.sendMsg( '/set', this.id, \extendedState, 1 );
//		_SCWindow_Minimize
//		^this.primitiveFailed
	}

	// JJJ begin
	unminimize {
		server.sendMsg( '/set', this.id, \extendedState, 0 );
	}
	// JJJ end

	alpha_ { arg alpha;
		// this would be perfect :
		// '/field', 'java.awt.SystemColor', \window
		// ... but : it's a texture which returns 0xFFFFFFFF thru getRGB ...
		// so we use grey ...
	
//		server.sendMsg( '/set', this.id, \background, *(Color( 1, 1, 1, alpha ).asSwingArg) );
		server.sendMsg( '/set', this.id, \background, *(Color( 0.8, 0.8, 0.8, alpha ).asSwingArg) );
		
//		_SCWindow_SetAlpha
//		^this.primitiveFailed
	}
	
	name_ { arg argName;
		name = argName;
		this.prSetName( argName );
	}
	
	bounds_ { arg argBounds;
		this.prSetBounds( argBounds );
	}
	
	setInnerExtent { arg w, h; // resize window keeping top left corner fixed
		var b;
		b = this.bounds;
		w = w ? b.width;
		if( h.isNil, { h = b.height }, { h = h + 45 }); // + 45 for window bar
		this.bounds = Rect.new( b.left, b.top + b.height - h, w, h );
	}
	
	bounds {
		^this.prGetBounds( Rect.new );
	}
	
	*screenBounds { arg server;
		^this.prGetScreenBounds( Rect.new, server );
	}
	
	play { arg function;
		AppClock.play({ 
			if( dataptr.notNil, {
				function.value;
			});
		});
	}
	
	findByID { arg id;
		^view.findByID( id );
	}

	prBoundsToJava { arg cocoa;
		var screenBounds;

		screenBounds 	= JSCWindow.screenBounds( server );

		// + 20 for window bar
		^Rect.new( cocoa.left, screenBounds.height - cocoa.top - cocoa.height - 22,
				  cocoa.width, cocoa.height + 22 );
	}
		
	// PRIVATE
	// primitives
	prInit { arg argName, argBounds, resizable, border; // , view;
		var viewID;

		bounds 	= argBounds;
		// tricky, we have to allocate the TopView's id here
		// to be able to assign our content pane to it, so
		// that JSCView can add key and dnd listeners
		viewID	= server.nextNodeID;

		acResp = OSCresponderNode( server.addr, "/window", { arg time, resp, msg;
			var state;
		
			case { msg[1] == this.id }
			{
				state = msg[2].asSymbol;
				case { (state === \resized) || (state === \moved) }
				{
					// XXX
				}
				{ state === \closing }
				{
					if( userCanClose, {
						{ this.prClose; }.defer;
					});
				}
			};
		}).add;

		server.sendBundle( nil,
			[ '/set', '[', "/local", this.id, '[', "/new", "de.sciss.swingosc.Frame", argName, ']', ']',
				\bounds ] ++ this.prBoundsToJava( argBounds ).asSwingArg ++ [ \resizable, resizable,
				\undecorated, border.not ],
			[ '/set', '[', "/local", viewID, '[', '/method', this.id, "getContentPane", ']', ']',
				\layout, '[', '/ref', \null, ']' ],
			[ "/local", "ac" ++ this.id,
				'[', "/new", "de.sciss.swingosc.WindowResponder", this.id, ']' ]
		);

		view = JSCTopView( this, argBounds.moveTo( 0, 0 ), server, viewID );

//		_SCWindow_New
//		^this.primitiveFailed
	}
	
	prClose {
		if( dataptr.notNil, {
			acResp.remove;
			this.drawHook_( nil );
			server.sendBundle( nil,
				[ '/method', "ac" ++ this.id, \remove ],
				[ '/method', this.id, \dispose ],
				[ "/free", "ac" ++ this.id, this.id ]);
			this.closed;
		},{
			"JSCWindow-remove : this view was already removed.".debug( this );
		});
//		_SCWindow_Close
//		^this.primitiveFailed
	}
	
	prSetName { arg argName;
		server.sendMsg( '/set', this.id, \title, argName.asSwingArg );
//		_SCWindow_SetName
//		^this.primitiveFailed
	}
	prGetBounds { arg argBounds;
		"JSCWindow.prGetBounds : not yet implemented".error;
		^argBounds.set( bounds.left, bounds.top, bounds.width, bounds.height );
//		_SCWindow_GetBounds
//		^this.primitiveFailed
	}
	prSetBounds { arg argBounds;
		bounds		= argBounds;
		argBounds		= this.prBoundsToJava( argBounds );

		server.listSendMsg([ '/set', this.id, \bounds ] ++ argBounds.asSwingArg );
//		_SCWindow_SetBounds
//		^this.primitiveFailed
	}

	prSetAcceptMouseOver{arg bool;
		"JSCWindow.acceptMouseOver : not yet implemented".error;
//		_SCWindow_SetAcceptMouseOver
//		^this.primitiveFailed		
	}

	*prGetScreenBounds { arg argBounds, server;
		server 		= server ?? { SwingOSC.default; };
		^argBounds.set( 0, 0, server.screenWidth, server.screenHeight );
//		_SCWindow_GetScreenBounds
//		^this.primitiveFailed
	}
	callDrawHook {
//		"JSCWindow.callDrawHook : not yet implemented".error;
//		drawHook.value( this );
		this.refresh;
	}
	
	id { ^id; }
	
	*viewPalette {
		var w, v, f, c;
		w = JSCWindow("View Palette", Rect(532, 64, 300, 320)); // .front; JJJ
		w.view.decorator = f = FlowLayout(w.view.bounds);
		
		c = [JSCSlider, JSCRangeSlider, JSC2DSlider, JSCPopUpMenu, JSCButton, 
			JSCNumberBox, JSCMultiSliderView,
			JSCStaticText, JSCDragSource, JSCDragSink, JSCDragBoth,
		];
		
		c.do({ arg item;
			var n;

			n = JSCDragSource(w, Rect(0, 0, 140, 24));
			n.object = item;
		
			item.paletteExample(w, Rect(0,0,140,24));
		});
		
		w.front;	// JJJ
	}
}
