/*
 *	JSCSoundFileView
 *	(SwingOSC classes for SuperCollider)
 *
 *	Replacement for the (Cocoa) SCSoundFileView class by Jan Truetzschler.
 *
 *	@author		Hanns Holger Rutz
 *	@version		0.441, 16-Dec-06
 */
JSCSoundFileView : JSCView { // JJJ not JSCScope
	var <soundfile;
	var <>metaAction, <elasticMode, <drawsWaveForm = true, <readProgress;
	var <startFrame, <numFrames;

	var <>dataFrames, <viewFrames, <>block=64;
	
	var viewStart	= 0;
	var sampleRate;
	var selections;
	
	var viewResp, asyncResp, asyncID;
	
	init { arg argParent, argBounds;
		super.init( argParent, argBounds );
		soundfile = SoundFile.new;
	}
	
	soundfile_ { arg snd;
		soundfile = snd;
		if( soundfile.isOpen ) { ^this };
		if( soundfile.openRead.not ) { ^this };
		soundfile.close;
	}

	read { arg startframe = 0, frames = 0, block = 64, closeFile = true;
		this.block = block;
		if( soundfile.isOpen.not ) {
			if( soundfile.openRead.not ) {
				^this;
			};
		};
		startFrame = startframe;
		numFrames = frames;
		this.readFile( soundfile, startframe, frames, block );
		if( closeFile) { soundfile.close };
	}

	readWithTask { arg startframe = 0, frames, block = 64, doneAction, showProgress = true;
		this.readFileWithTask( soundfile, startframe, frames, block, doneAction, showProgress );
	}
	
	makeProgressWindow {
		this.addDependant( JSoundFileViewProgressWindow(soundfile.path.split.last ));
	}
	
	// needs an open soundfile;
	readFileWithTask { arg soundfile, startframe = 0, frames, block = 64, 
						doneAction, showProgress = true;
		this.block = block;
		if( showProgress ) {		
			this.makeProgressWindow;
		};
		frames 		= frames ? soundfile.numFrames;
		sampleRate	= soundfile.sampleRate;
		{
			this.prCallRead( soundfile, startFrame, frames, block,
						   soundfile.sampleRate, soundfile.numFrames, soundfile.numChannels );
			viewStart		= 0;
			dataFrames	= this.dataNumSamples / this.soundfile.numChannels;
			doneAction.value( this );
			soundfile.close;
		}.fork( AppClock );		
	}

	readFile{ arg asoundfile, startframe = 0, frames = 0, block = 0, closefile = true;
		sampleRate = asoundfile.sampleRate;
		if( thisThread.isKindOf( Routine )) {
			^this.prCallRead( asoundfile, startframe, frames, block, asoundfile.sampleRate, 
							asoundfile.numFrames, asoundfile.numChannels );
		} {
			"JSCSoundFileView.read : asynchronous call outside routine".warn;
			{
				this.prCallRead( asoundfile, startframe, frames, block, asoundfile.sampleRate, 
							   asoundfile.numFrames, asoundfile.numChannels
				);
			}.fork( AppClock );
		};
		viewStart		= 0;
		dataFrames	= this.dataNumSamples / this.soundfile.numChannels;
	}

//	mouseEndTrack{|x,y|
////		mouseUpAction.value(this, x,y)
//	}
	
	// ??? what is this for actually ???
	doMetaAction {
		metaAction.value( this);
	}
	
	currentSelection_ { arg index;
		this.setProperty( \selectionIndex, index );
	}
	
	currentSelection {
		^this.getProperty( \selectionIndex );	
	}
	
	setSelectionStart { arg index, frame;
		var sel;
		sel = selections[ index ];
		if( sel.notNil ) {
			sel.put( \start, frame.max( 0 ));
			this.prSendSelection( index );
		} {
			this.prMethodError( thisMethod, "Illegal selection index "++index );
		};
	}

	setSelectionSize { arg index, frame;
		var sel;
		sel = selections[ index ];
		if( sel.notNil ) {
			sel.put( \size, frame.max( 0 ));
			this.prSendSelection( index );
		} {
			this.prMethodError( thisMethod, "Illegal selection index "++index );
		};
	}
	
	// extra function in SwingOSC
	setSelectionSpan { arg index, startFrame, numFrames;
		var sel;
		sel = selections[ index ];
		if( sel.notNil ) {
			sel.put( \start, startFrame.max( 0 ));
			sel.put( \size, numFrames.max( 0 ));
			this.prSendSelection( index );
		} {
			this.prMethodError( thisMethod, "Illegal selection index "++index );
		};
	}
	
	setEditableSelectionStart { arg index, bool;
		server.sendMsg( '/method', this.id, \setSelectionStartEditable, index, bool );
	}
	
	setEditableSelectionSize { arg index, bool;
		server.sendMsg( '/method', this.id, \setSelectionSizeEditable, index, bool );
	}
		
	setSelectionColor { arg index, color;
//		this.setProperty( \selectionColor, [ index, color ]);
		server.listSendMsg([ '/method', this.id, \setSelectionColor, index ] ++ color.asSwingArg );
	}
	
	selections {
		^selections.collect({ arg sel; [ sel[ \start ] ?? 0, sel[ \size ] ?? 0 ]});
	}
	
	selectionStart { arg index;
		var sel;
		sel = selections[ index ];
		if( sel.notNil ) {
			^sel[ \start ];
		} {
			this.prMethodError( thisMethod, "Illegal selection index "++index );
			^nil;
		};
	}

	selectionSize { arg index;
		var sel;
		sel = selections[ index ];
		if( sel.notNil ) {
			^sel[ \size ];
		} {
			this.prMethodError( thisMethod, "Illegal selection index "++index );
			^nil;
		};
	}

	selectionStartTime { arg index;
		var frame;
		frame = this.selectionStart( index );
		if( frame.notNil ) {
			^frame / (sampleRate ?? 44100);
		} {
			^nil;
		};
	}

	selectionDuration { arg index;
		var frames;
		frames = this.selectionSize( index );
		if( frames.notNil ) {
			^frames / (sampleRate ?? 44100);
		} {
			^nil;
		};
	}
	
	readSelection {
		this.read(
			this.selectionStart(this.currentSelection),
			this.selectionSize(this.currentSelection)
		);
	}
	
	readSelectionWithTask {
		this.readWithTask(
			this.selectionStart(this.currentSelection),
			this.selectionSize(this.currentSelection)
		);
	}
		
	gridOn_ { arg boolean;
		this.setProperty( \gridPainted, boolean );
	}
	
	gridResolution_ { arg resolution;
		this.setProperty( \gridResolution, resolution );
	}
	
	dataNumSamples {
		^(soundfile.numFrames * soundfile.numChannels).div( block );
	}
	
	data {
		var arr;
		arr = Array.newClear( this.dataNumSamples );
		^this.getProperty( \getViewData, arr );
	}
	
	data_ { arg arr;
		this.setData( arr, 64, 0, soundfile.numChannels, soundfile.sampleRate );
	}
	
	setData { arg arr, block = 64, startframe = 0, channels = 1, samplerate = 44100;
this.prMethodError( thisMethod, "Not yet implemented" );
		this.block = block;
		this.setProperty( \setViewData, [ arr, block, startframe, channels, samplerate ]);
	}
		
	elasticMode_{arg mode;
if( mode.not, { "JSCSoundFileView.elasticMode_( false ) not yet supported!".warn });
		elasticMode = mode;
//		this.setProperty( \elasticResizeMode, mode );
	}	
	
	drawsWaveForm_ { arg bool;
		drawsWaveForm = bool;
		this.setProperty( \wavePainted, bool );
	}
	
	timeCursorPosition_ { arg frame;
		this.setProperty( \timeCursorPosition, frame );
	}
	
	timeCursorPosition {
		^this.getProperty( \timeCursorPosition );
	}
	
	timeCursorOn_ { arg bool;
		this.setProperty( \timeCursorPainted, bool );
	}
	
	timeCursorColor_ { arg color;
		this.setProperty( \timeCursorColor, color );
	}

	zoom { arg factor;	// zoom factor n or 1/n.
		viewFrames	= (viewFrames * factor + 0.5).asInteger.clip( 0, soundfile.numFrames );
		viewStart		= viewStart.min( soundfile.numFrames - viewFrames );
		this.prUpdateViewSpan;
	}

	zoomToFrac { arg frac; // 0..1
		viewFrames 	= (soundfile.numFrames * frac.clip( 0, 1 ) + 0.5).asInteger;
		viewStart		= viewStart.min( soundfile.numFrames - viewFrames );
		this.prUpdateViewSpan;
	}

	zoomAllOut {
		viewStart		= 0;
		viewFrames	= soundfile.numFrames;
		this.prUpdateViewSpan;
	}
	
	zoomSelection { arg index;	// selection index
		var sel;
		sel = selections[ index ];
		if( this.selectionSize( index) > 0 ) {
			this.x_( this.selectionStart( index ));
			this.xZoom = this.selectionSize( index ) / (this.bounds.width - 2) / block;
		};
	}
	
	scrollTo { arg position;		// absolute. from 0 to 1
		viewStart 	= ((soundfile.numFrames - viewFrames) * position.clip( 0, 1 ) + 0.5).asInteger;
		this.prUpdateViewSpan;
	}
	
	scroll { arg amount;	// +/- range in viewFrames
		viewStart 	= (viewStart + (amount * viewFrames)).clip( 0, soundfile.numFrames - viewFrames );
		this.prUpdateViewSpan;
	}
	
	scrollToStart {
		this.x_( 0 );
	}
	
	scrollToEnd {
		viewStart = soundfile.numFrames  - viewFrames;
		this.prUpdateViewSpan;
	}
	
	selectAll { arg index;	// selection index
		this.setSelectionSpan( index, 0, this.soundfile.numFrames );
	}
	
	selectNone { arg index;	 // selection index
		 // sends x to 0 when selection is not full width visible? ( thus update scrollPos )		this.setSelectionSize( index, 0 );
	}
	
	scrollPos {
		^viewStart / (soundfile.numFrames - viewFrames).max( 1 );
	}
	
	gridOffset_{|offset|
		this.setProperty(\gridOffset, offset);		
	}

	// (these are taken from JSCScope since we don't subclass that)
	x {
		^viewStart;
	}

	x_ { arg val;
		viewStart 	= val.min( val, soundfile.numFrames - viewFrames );
		viewFrames	= viewFrames.min( viewFrames, soundfile.numFrames - viewStart );
		this.prUpdateViewSpan;
	}	

//	y {
//		^this.getProperty(\y)
//	}
//	y_ { arg val;
//		this.setProperty(\y, val);
//	}	

	xZoom {
		^( viewFrames / ((this.bounds.width - 2) * block) );
	}

	xZoom_ { arg val;
		viewFrames 	= (block * val * (this.bounds.width - 2)).clip( 0, soundfile.numFrames );
		viewStart 	= viewStart.min( soundfile.numFrames - viewFrames );
		this.prUpdateViewSpan;
	}	

	yZoom {
		^this.getProperty( \yZoom );
	}

	yZoom_ { arg val;
		this.setProperty( \yZoom, val );
	}	

	gridColor {
		^this.getProperty( \gridColor );
	}

	gridColor_ { arg color;
		this.setProperty( \gridColor, color );
	}	

	waveColors {
		^this.getProperty( \objWaveColors );
	}
	
	waveColors_ { arg arrayOfColors;
		this.setProperty( \objWaveColors, arrayOfColors );
	}
	
	style_ { arg val;
if( val == 2, {"JSCSoundFileView.style_ : lissajou not yet implemented".error; ^this; }); // XXX
		this.setProperty( \style, val );
		// 0 = vertically spaced
		// 1 = overlapped
		// 2 = x/y
	}
	
//	properties {
//		^super.properties ++ #[ \yZoom, \gridColor, \waveColors, \style ]
//	}

	prSCViewNew {
//		properties.put( \bufnum, 0 );
//		properties.put( \x, 0.0 );
//		properties.put( \y, 0.0 );
//		properties.put( \xZoom, 1.0 );
		properties.put( \yZoom, 1.0 );
		properties.put( \style, 0 );
		
		selections = Array.fill( 64, {IdentityDictionary.new });
		
		viewResp = OSCresponderNode( server.addr, '/soundfile', { arg time, resp, msg;
			var state, sel, idx, frame, size;
		
			if( msg[1] === this.id, {
				state = msg[2].asSymbol;
				case
				{ state === \cursor }
				{
					frame = msg[3];
					properties.put( \timeCursorPosition, frame );
					{action.value( this, \cursor, frame )}.defer;
				}
				{ state === \selection }
				{
					idx = msg[3];
					sel = selections[ idx ];
					if( sel.notNil ) {
						frame = msg[4];
						sel.put( \start, frame );
						size = msg[5] - frame;
						sel.put( \size, size );
						{action.value( this, \selection, idx, frame, size )}.defer;
					} {
						"Yukk! Selection not found".error;
					};
				};
			});
		});
		viewResp.add;

		^super.prSCViewNew([
			[ '/local', this.id, '[', '/new', "de.sciss.swingosc.SoundFileView", ']',
				"vw" ++ this.id,
				'[', '/new', "de.sciss.swingosc.SoundFileViewResponder", this.id, ']' ]
			]);
	}

	prClose {
		viewResp.remove;
		this.prFreeAsyncResp;
		^super.prClose([[ '/method', "vw" ++ this.id, \remove ],
					   [ '/free', "vw" ++ this.id ],
					   [ '/method', this.id, \dispose ]]);
	}

//	// JJJ begin
//	prSendProperty { arg key, value;
//		key	= key.asSymbol;
//
//		switch( key,
//			\readSndFile, {
//				if( thisThread.isKindOf( Routine ), {
//					^this.prCallRead( *value );
//				}, {
//					"JSCSoundFileView.read : asynchronous call outside routine".warn;
//					fork { this.prCallRead( *value )};
//				});
//				^this;
//			}
//		);
//		^super.prSendProperty( key, value );
//	}
//	// JJJ end

/* private methods*/

// JJJ begin
//	updateScroll {
//		scrollPos = this.x / (soundfile.numFrames - viewFrames)
//	}
// JJJ end
	
// JJJ begin
//	updateData {
//		scrollPos = 0;
//		dataFrames = this.dataNumSamples/this.soundfile.numChannels;
//		zoomOne = dataFrames / (this.bounds.width-2);
//		viewFrames = dataFrames * block * (this.xZoom / zoomOne);
//	}
// JJJ end

	prCallRead {arg sf, startFrame, numFrames, block, sfSampleRate, sfNumFrames, sfNumChannels;
		var condition, bndl, cancel, timeout;
		condition = Condition.new;
		
		this.prFreeAsyncResp;

		asyncResp	= OSCresponderNode( server.addr, '/async', { arg time, resp, msg;
			var what;

			if( msg[ 1 ] == this.id, {
				if( cancel.notNil, {cancel.stop; });
				what = msg[ 2 ];
				case
				{ what === \finished }
				{
					this.prFreeAsyncResp;
//"JSCSoundFileView.read : done".inform;
					condition.test	= true;
					condition.signal;
					this.changed( \progressFinished );
				}
				{ what === \failed }
				{
					this.prFreeAsyncResp;
					"JSCSoundFileView.read : failed".error;
					condition.test	= true;
					condition.signal;
					this.changed( \progressFinished );
				}
				{ what === \cancelled }
				{
					this.prFreeAsyncResp;
					condition.test	= true;
					condition.signal;
					this.changed( \progressFinished );
				}
				{ what === \update }
				{
					readProgress = msg[ 4 ];
					this.changed( \progress );
				};
			});
		});
		asyncResp.add;
		
		asyncID = "as" ++ server.nextNodeID;
		bndl = List.new;
		bndl.add([ '/method', this.id, \cancelAsyncRead ]);
		bndl.add([ '/local', asyncID, '[', '/new', "de.sciss.swingosc.AsyncResponder", this.id, \readProgress, ']' ]);
		bndl.add([ '/method', this.id, \readSndFile, sf.path, startFrame, numFrames ]);

		condition.test = false;
		cancel = fork {
			8.0.wait;
			asyncResp.remove;
			"JSCSoundFileView.read : timeout".error;
			condition.test	= true;
			condition.signal;
		};
		readProgress = 0;
		server.listSendBundle( nil, bndl );
		this.changed( \progress );
		condition.wait;
	}
	
	prFreeAsyncResp {
		if( asyncResp.notNil, {
			asyncResp.remove;
			asyncResp = nil;
			server.sendBundle( nil, [ '/method', asyncID, \remove ], [ '/free', asyncID ]);
			asyncID = nil;
		});
	}	

	prUpdateViewSpan {
		server.sendMsg( '/method', this.id, \setViewSpan, viewStart, viewStart + viewFrames );
	}

	prMethodError {arg methodName, message;
		(this.class.name ++ "." ++ methodName ++ " failed : " ++ message).error;
	}
	
	prSendSelection { arg index;
		var sel, start, stop;
		sel 		= selections[ index ];
		start	= sel[ \start ] ?? 0;
		stop		= start + (sel[ \size ] ?? 0);
		server.sendMsg( '/method', this.id, \setSelectionSpan, index, start, stop );
	}

	prBoundsToJava { arg rect;
		var pb;
		
		pb          = parent.bounds;
		// focus-border must be painted inside gadget bounds
		^rect.insetAll( -3, -3, -3, -3 ).moveBy( pb.left.neg, pb.top.neg );
	}

	prBoundsFromJava { arg rect;
		var pb;
		
		pb          = parent.bounds;
		^rect.moveBy( pb.left, pb.top ).insetAll( 3, 3, 3, 3 );
	}
}

// note: this should be killed in favour of
// making SoundFileViewProgressWindow use the GUI class
JSoundFileViewProgressWindow{
	var win, slider;
	
	// JJJ begin
	var popUpRout;
	// JJJ end
	
	*new{|name|
		^super.new.makeWindow(name)
	}
	
	makeWindow{|name|
		// JJJ begin
		win = JSCWindow.new( "Reading: " ++ name,
				 Rect( 100, 100, 300, 40 ), false );
		win.view.decorator = FlowLayout(win.view.bounds);
		slider = JSCRangeSlider(win, Rect(4,4,290,10))
//			.editable_( false )
			.canFocus_( false );
		popUpRout = {1.5.wait; win.front; }.fork( AppClock );
		// JJJ end
	}
	
	update{|changed, changer|
		
		if(changer === \progress){
			{slider.lo_(0).hi_(changed.readProgress)}.defer
			^this
		};
		// JJJ begin
		if( changer === \progressFinished, {
			popUpRout.stop;
			{ win.close;
			  changed.removeDependant( this);
			}.defer;
		});
		// JJJ end;
	}
}